"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DialectPgsql = void 0;
var tslib_1 = require("tslib");
var builder_1 = require("@typed-query-builder/builder");
var sql_1 = require("@typed-query-builder/sql");
require("./types");
var PostgresDialect = /** @class */ (function (_super) {
    tslib_1.__extends(PostgresDialect, _super);
    function PostgresDialect() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.gis = false;
        _this.colonCasting = false;
        return _this;
    }
    return PostgresDialect;
}(sql_1.Dialect));
exports.DialectPgsql = new PostgresDialect();
exports.DialectPgsql.removeSupport([
    sql_1.DialectFeatures.UNSIGNED,
    sql_1.DialectFeatures.NAMED_PARAMETERS,
]);
exports.DialectPgsql.addReservedWords(sql_1.ReservedWords);
exports.DialectPgsql.implicitPredicates = true;
exports.DialectPgsql.aliasQuotesOptional = /^[a-z\d_]+$/;
exports.DialectPgsql.nameQuotesOptional = /^[a-z\d_]+$/;
exports.DialectPgsql.predicateBinary.alias('!=', '<>');
exports.DialectPgsql.predicateBinaryList.alias('!=', '<>');
exports.DialectPgsql.operationBinary.aliases({
    BITAND: '&',
    BITOR: '|',
    BITXOR: '^',
    BITLEFT: '<<',
    BITRIGHT: '>>',
});
exports.DialectPgsql.operationUnary.alias('BITNOT', '~');
exports.DialectPgsql.predicateUnary.setFormats({
    FALSE: '{value} IS FALSE',
    TRUE: '{value}',
});
exports.DialectPgsql.joinType.aliases({
    INNER: 'INNER JOIN',
    LEFT: 'LEFT JOIN',
    RIGHT: 'RIGHT JOIN',
    FULL: 'FULL JOIN',
});
exports.DialectPgsql.lateralJoinType.aliases({
    INNER: 'INNER JOIN LATERAL',
    LEFT: 'LEFT JOIN LATERAL',
    RIGHT: 'RIGHT JOIN LATERAL',
    FULL: 'FULL JOIN LATERAL',
});
// =========================================================
// Data Types
// =========================================================
exports.DialectPgsql.setDataTypeUnsupported('LINE');
exports.DialectPgsql.valueFormatterMap.BOOLEAN = sql_1.Dialect.FormatBoolean;
exports.DialectPgsql.setDataTypeFormat('BIT', { constant: 'BIT(1)' });
exports.DialectPgsql.setDataTypeFormat('BITS', { constant: 'BITS(1)', tuple: 'BITS({1})' });
exports.DialectPgsql.setDataTypeFormat('MEDIUMINT', { constant: 'INT', tuple: 'INT' });
exports.DialectPgsql.setDataTypeFormat('TINYINT', { constant: 'SMALLINT', tuple: 'SMALLINT' });
exports.DialectPgsql.valueFormatterMap.BIT = formatInteger;
exports.DialectPgsql.valueFormatterMap.TINYINT = formatInteger;
exports.DialectPgsql.valueFormatterMap.SMALLINT = formatInteger;
exports.DialectPgsql.valueFormatterMap.MEDIUMINT = formatInteger;
exports.DialectPgsql.valueFormatterMap.INT = formatInteger;
exports.DialectPgsql.valueFormatterMap.BIGINT = formatInteger;
exports.DialectPgsql.valueFormatterMap.BITS = formatInteger;
exports.DialectPgsql.setDataTypeFormat('FLOAT', { constant: 'REAL' });
exports.DialectPgsql.valueFormatterMap.FLOAT = formatFloat;
exports.DialectPgsql.setDataTypeFormat('DOUBLE', { constant: 'DOUBLE PRECISION' });
exports.DialectPgsql.valueFormatterMap.DOUBLE = formatFloat;
exports.DialectPgsql.valueFormatterMap.NUMERIC = formatDecimal;
exports.DialectPgsql.valueFormatterMap.DECIMAL = formatDecimal;
exports.DialectPgsql.valueFormatterMap.MONEY = formatDecimal;
exports.DialectPgsql.valueFormatterMap.CHAR = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.VARCHAR = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.TEXT = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.TIMESTAMP = function (v, d, t) { return d.quoteValue(v.text) + '::timestamp' + (v.hasTimeZone ? 'tz' : ''); };
exports.DialectPgsql.valueFormatterMap.DATE = function (v, d, t) { return d.quoteValue(v.text) + '::date'; };
exports.DialectPgsql.valueFormatterMap.TIME = function (v, d, t) { return d.quoteValue(v.text) + '::time' + (v.hasTimeZone ? 'tz' : ''); };
exports.DialectPgsql.valueFormatterMap.UUID = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.CIDR = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.INET = sql_1.Dialect.FormatString;
exports.DialectPgsql.valueFormatterMap.MACADDR = sql_1.Dialect.FormatString;
exports.DialectPgsql.setDataTypeFormat('BINARY', { constant: 'BYTEA' });
exports.DialectPgsql.valueFormatterMap.BINARY = formatBinary;
exports.DialectPgsql.setDataTypeFormat('VARBINARY', { constant: 'BYTEA' });
exports.DialectPgsql.valueFormatterMap.VARBINARY = formatBinary;
exports.DialectPgsql.setDataTypeFormat('BLOB', { constant: 'BYTEA', tuple: 'BYTEA' });
exports.DialectPgsql.valueFormatterMap.BLOB = formatBinary;
exports.DialectPgsql.valueFormatterMap.JSON = function (v, d) { return d.quoteValue(JSON.stringify(v)) + '::json'; };
exports.DialectPgsql.valueFormatterMap.XML = function (v, d) { return d.quoteValue(String(v)) + '::xml'; };
exports.DialectPgsql.valueFormatterMap.INTERVAL = function (v, d) {
    if (v instanceof builder_1.DataInterval) {
        return 'interval ' + d.quoteValue(v.toString());
    }
    var i = builder_1.DataInterval.from(v);
    if (!i.isValid()) {
        return String(v);
    }
    else {
        return 'interval ' + d.quoteValue(i.toString());
    }
};
exports.DialectPgsql.valueFormatterMap.POINT = function (v, d) {
    return formatGeometry(v, d, function (g, s) { return "ST_Point(" + g(s.x) + ", " + g(s.y) + ")"; }, function (g, s) { return "point(" + g(s.x) + ", " + g(s.y) + ")"; });
};
exports.DialectPgsql.setDataTypeFormat('SEGMENT', { constant: 'LSEG' });
exports.DialectPgsql.valueFormatterMap.SEGMENT = function (v, d, t) {
    return formatGeometry(v, d, function (g, s) { return "ST_MakeLine(ST_Point(" + g(s.x1) + ", " + g(s.y1) + "), ST_Point(" + g(s.x2) + ", " + g(s.y2) + "))"; }, function (g, s) { return "lseg('[(" + g(s.x1, true) + ", " + g(s.y1, true) + "), (" + g(s.x2, true) + ", " + g(s.y2, true) + ")]')"; });
};
exports.DialectPgsql.valueFormatterMap.CIRCLE = function (v, d, t) {
    return formatGeometry(v, d, function (g, s) { return "ST_Buffer(ST_Point(" + g(s.x) + ", " + g(s.y) + "), " + g(s.r) + ", 'quad_segs=32')"; }, function (g, s) { return "circle('<(" + g(s.x, true) + ", " + g(s.y, true) + "), " + g(s.r, true) + ">')"; });
};
exports.DialectPgsql.valueFormatterMap.PATH = function (v, d, t) {
    return formatGeometry(v, d, function (g, s) { return "ST_MakeLine(ARRAY[" + s.points.map(function (_a) {
        var x = _a.x, y = _a.y;
        return "ST_Point(" + g(x) + ", " + g(y) + ")";
    }).join(', ') + "])"; }, function (g, s) { return "path('[" + s.points.map(function (_a) {
        var x = _a.x, y = _a.y;
        return "(" + g(x, true) + ", " + g(y, true) + ")";
    }).join(', ') + "]')"; });
};
exports.DialectPgsql.valueFormatterMap.POLYGON = function (v, d, t) {
    return formatGeometry(v, d, function (g, s) { return "ST_MakePolygon(ST_MakeLine(ARRAY[" + s.corners.map(function (_a) {
        var x = _a.x, y = _a.y;
        return "ST_Point(" + g(x) + ", " + g(y) + ")";
    }).join(', ') + "]))"; }, function (g, s) { return "polygon('(" + s.corners.map(function (_a) {
        var x = _a.x, y = _a.y;
        return "(" + g(x, true) + ", " + g(y, true) + ")";
    }).join(', ') + ")')"; });
};
exports.DialectPgsql.valueFormatterMap.BOX = function (v, d, i) {
    return formatGeometry(v, d, function (g, s) { return "ST_MakePolygon(ST_MakeLine(ARRAY[ ST_Point(" + g(s.minX) + ", " + g(s.minY) + "), ST_Point(" + g(s.maxX) + ", " + g(s.minY) + "), ST_Point(" + g(s.maxX) + ", " + g(s.maxY) + "), ST_Point(" + g(s.minX) + ", " + g(s.maxY) + ") ]))"; }, function (g, s) { return "box('((" + g(s.minX, true) + ", " + g(s.minY, true) + "), (" + g(s.maxX, true) + ", " + g(s.maxY, true) + "))')"; });
};
exports.DialectPgsql.dataTypeFormatter.POINT = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.PATH = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.POLYGON = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.LINE = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.SEGMENT = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.BOX = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.CIRCLE = formatGeometryType;
exports.DialectPgsql.dataTypeFormatter.GEOGRAPHY = function (dataType) {
    return dataType === 'GEOGRAPHY' || dataType[0] === 'GEOGRAPHY'
        ? 'GEOGRAPHY'
        : dataType[1] && dataType[1] !== 4326
            ? "GEOGRAPHY(" + dataType[0] + ", " + dataType[1] + ")"
            : "GEOGRAPHY(" + dataType[0] + ")";
};
function formatDecimal(value, dialect, dataType) {
    return value.toFixed(builder_1.getDataTypeMeta(dataType).fractionDigits || 0);
}
function formatInteger(value) {
    return value.toFixed(0);
}
function formatFloat(value) {
    return value.toString();
}
function formatBinary(value, dialect) {
    if (builder_1.isString(value)) {
        return 'E' + dialect.quoteValue('\\\\x' + value.split('').map(escapeBinary).join(''));
    }
    if (value instanceof Buffer) {
        var hex = value.toString('hex');
        return 'E' + dialect.quoteValue('\\\\x' + hex);
    }
    return value;
}
function formatGeometry(value, dialect, getGis, getNormal) {
    var source = value.deep || value;
    var get = value.deep
        ? function (value, inQuote) {
            if (inQuote === void 0) { inQuote = false; }
            if (value instanceof builder_1.Expr) {
                var query = dialect.output()(value).query;
                return inQuote
                    ? "'||" + query + "||'"
                    : query;
            }
            else {
                return String(value);
            }
        }
        : function (value, inQuote) {
            if (inQuote === void 0) { inQuote = false; }
            return builder_1.isNumber(value) || !inQuote
                ? String(value)
                : "'||" + value + "||'";
        };
    return exports.DialectPgsql.gis
        ? value.srid
            ? "ST_SetSRID(" + getGis(get, source) + ", " + value.srid + ")"
            : getGis(get, source)
        : getNormal(get, source);
}
function formatGeometryType(type) {
    if (!exports.DialectPgsql.gis) {
        return type;
    }
    return typeof type === 'string' || !Array.isArray(type)
        ? type
        : typeof type[1] !== 'number' || type[1] === 4326
            ? "GEOGRAPHY(" + type[0] + ")"
            : "GEOGRAPHY(" + type[0] + ", " + type[1] + ")";
}
function escapeBinary(x) {
    if (x === '\\') {
        return '\\\\';
    }
    var code = x.codePointAt(0);
    if (code <= 31 || code >= 127) {
        var octal = code.toString(8);
        return '\\000'.substring(0, 4 - octal.length) + octal;
    }
    return x;
}
// Fallbacks
exports.DialectPgsql.valueFormatter.push(sql_1.Dialect.FormatBoolean, sql_1.Dialect.FormatDate, sql_1.Dialect.FormatString, sql_1.Dialect.FormatNumber, sql_1.Dialect.FormatNull);
// =========================================================
// Functions
// =========================================================
exports.DialectPgsql.functions.aliases({
    minScale: 'MIN_SCALE',
    trimScale: 'TRIM_SCALE',
    truncate: 'TRUNC',
    widthBucket: 'WIDTH_BUCKET',
    bitLength: 'BIT_LENGTH',
    toHex: 'TO_HEX',
    octetLength: 'OCTET_LENGTH',
    getBit: 'GET_BIT',
    getByte: 'GET_BYTE',
    setBit: 'SET_BIT',
    setByte: 'SET_BYTE',
    trimLeft: 'LTRIM',
    trimRight: 'RTRIM',
    padLeft: 'LPAD',
    padRight: 'RPAD',
    indexOf: 'STRPOS',
    substring: 'SUBSTR',
    startsWith: 'STARTS_WITH',
    regexReplace: 'REGEXP_REPLACE',
    char: 'CHR',
    join: 'CONCAT_WS',
    dateParse: 'TO_DATE',
    timestampParse: 'TO_TIMESTAMP',
    dateFormat: 'TO_CHAR',
    dateGet: 'DATE_PART',
    dateTruncate: 'DATE_TRUNC',
    createDate: 'MAKE_DATE',
    createTime: 'MAKE_TIME',
    createTimestamp: 'MAKE_TIMESTAMP',
    timestampFromSeconds: 'TO_TIMESTAMP',
    uuid: 'GEN_RANDOM_UUID',
    getAge: 'AGE',
    intervalGet: 'DATE_PART',
    intervalTruncate: 'DATE_TRUNC',
    createInterval: 'MAKE_INTERVAL',
});
exports.DialectPgsql.functions.setFormats({
    regexGet: 'REGEXP_MATCH({0}, {1})[1]',
    currentTime: 'CURRENT_TIME',
    currentTimestamp: 'CURRENT_TIMESTAMP',
    currentDate: 'CURRENT_DATE',
    iif: '(CASE WHEN {0} THEN {1} ELSE {2} END)',
    dateAdd: "({2} + interval '{1} {0}')",
    dateDiff: "(date_part({0}, {2}) - date_part({0}, {1}))",
    timestampToSeconds: 'EXTRACT(EPOCH FROM {0})',
    datesOverlap: '(({0}, {1}) OVERLAPS ({2}, {3}))',
    timestampsOverlap: '(({0}, {1}) OVERLAPS ({2}, {3}))',
    geomTranslate: '(({0})+({1}))',
    geomPathConcat: '(({0})+({1}))',
    geomDivide: '(({0})/({1}))',
    geomSame: '(({0})~=({1}))',
    dateAddInterval: '({0}+{1})',
    dateSubInterval: '({0}-{1})',
    timestampAddInterval: '({0}+{1})',
    timestampSubInterval: '({0}-{1})',
    timestampInterval: '({0}-{1})',
    timeSubInterval: '({0}-{1})',
    timeInterval: '({0}-{1})',
    intervalAdd: '({0}+{1})',
    intervalSub: '({0}-{1})',
    intervalNegate: '(-{0})',
    intervalMultiply: '(({0})*({1}))',
    intervalDivide: '(({0})/({1}))',
});
exports.DialectPgsql.functions.formats.geomContains = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Contains(" + a + ", " + b + ")"
        : "((" + a + ")@>(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomDistance = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Distance(" + a + ", " + b + ")"
        : "((" + a + ")<->(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomIntersection = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Intersection(" + a + ", " + b + ")"
        : "((" + a + ")#(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomIntersects = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Intersects(" + a + ", " + b + ")"
        : "((" + a + ")?#(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomTouches = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Touches(" + a + ", " + b + ")"
        : "((" + a + ")&&(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomWithinDistance = function (_a) {
    var _b = tslib_1.__read(_a, 3), a = _b[0], b = _b[1], c = _b[2];
    return exports.DialectPgsql.gis
        ? "ST_DWithin(" + a + ", " + b + ", " + c + ")"
        : "((" + a + ")<->(" + b + ")) <= " + c;
};
exports.DialectPgsql.functions.formats.geomScale = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_Scale(" + a + ", " + b + ")"
        : "((" + a + ")*(" + b + "))";
};
exports.DialectPgsql.functions.formats.geomPointX = function (_a) {
    var _b = tslib_1.__read(_a, 1), a = _b[0];
    return exports.DialectPgsql.gis
        ? "ST_X(" + a + ")"
        : a + "[0]";
};
exports.DialectPgsql.functions.formats.geomPointY = function (_a) {
    var _b = tslib_1.__read(_a, 1), a = _b[0];
    return exports.DialectPgsql.gis
        ? "ST_Y(" + a + ")"
        : a + "[1]";
};
exports.DialectPgsql.functions.formats.geomPoints = function (_a) {
    var _b = tslib_1.__read(_a, 1), a = _b[0];
    return exports.DialectPgsql.gis
        ? "ST_NPoints(" + a + ")"
        : "NPOINTS(" + a + ")";
};
exports.DialectPgsql.functions.formats.geomPoint = function (_a) {
    var _b = tslib_1.__read(_a, 2), a = _b[0], b = _b[1];
    return exports.DialectPgsql.gis
        ? "ST_PointN(" + a + ", " + b + ")"
        : "ST_PointN(" + a + ", " + b + ")";
};
exports.DialectPgsql.functions.formats.geomLength = function (_a) {
    var _b = tslib_1.__read(_a, 1), a = _b[0];
    return exports.DialectPgsql.gis
        ? "ST_Length(" + a + ")"
        : "LENGTH(" + a + ")";
};
exports.DialectPgsql.functions.formats.geomCenter = function (_a) {
    var _b = tslib_1.__read(_a, 1), a = _b[0];
    return exports.DialectPgsql.gis
        ? "ST_Centroid(" + a + ")"
        : "CENTER(" + a + ")";
};
exports.DialectPgsql.aggregates.aliases({
    bitAnd: 'BIT_AND',
    bitOr: 'BIT_OR',
    boolAnd: 'BOOL_AND',
    boolOr: 'BOOL_OR',
    string: 'STRING_AGG',
    rowNumber: 'ROW_NUMBER',
    denseRank: 'DENSE_RANK',
    percentRank: 'PERCENT_RANK',
    cumulativeDistribution: 'CUME_DIST',
    firstValue: 'FIRST_VALUE',
    lastValue: 'LAST_VALUE',
    nthValue: 'NTH_VALUE',
    deviation: 'STDDEV',
});
exports.DialectPgsql.aggregates.setFormats({
    countIf: 'COUNT({distinct}CASE WHEN ({0}) = 1 THEN 1 ELSE NULL END){over}',
});
exports.DialectPgsql.functions.setCascadings({
    random: [
        [1, '(random() * (({0}) - ({1})) + ({1}))'],
        [0, '(random() * ({0}))'],
        ['*', 'random()'],
    ],
});
exports.DialectPgsql.functionsRawArguments.dateAdd = { 0: true };
// =========================================================
// Expressions
// =========================================================
sql_1.addExprs(exports.DialectPgsql);
sql_1.addFeatures(exports.DialectPgsql);
sql_1.addQuery(exports.DialectPgsql);
sql_1.addSources(exports.DialectPgsql);
exports.DialectPgsql.transformer.setTransformer(builder_1.QueryJson, function (expr, transform, out) {
    var json = expr.json;
    var subquery = out.modify({ includeSelectAlias: true }, function () { return transform(json, out); });
    if (json instanceof builder_1.QueryFirst) {
        return "SELECT row_to_json(t) FROM (" + subquery + ") as t";
    }
    else if (json instanceof builder_1.QuerySelect) {
        return "SELECT coalesce(json_agg(row_to_json(t)), '[]'::json) FROM (" + subquery + ") as t";
    }
    else if (json instanceof builder_1.QueryList) {
        return "SELECT coalesce(json_agg(t.item), '[]'::json) FROM (" + subquery + ") as t";
    }
    else {
        throw new Error('Converting the request operation to JSON is not supported.');
    }
});
exports.DialectPgsql.transformer.setTransformer(builder_1.ExprCast, function (expr, transform, out) {
    var value = expr.value, type = expr.type;
    var x = '';
    var v = out.wrap(value);
    var t = out.dialect.getDataTypeString(type);
    if (exports.DialectPgsql.colonCasting) {
        x += v;
        x += '::';
        x += t;
    }
    else {
        x += 'CAST(';
        x += v;
        x += ' AS ';
        x += t;
        x += ')';
    }
    return x;
});
//# sourceMappingURL=index.js.map